import os
import pkg_resources
from fk.utils import get_package_file_content

import logging

logger = logging.getLogger(__name__)

version_filename = "VERSION"

__version__ = get_package_file_content(version_filename) or "0.0.0"



import psycopg2
import logging

from .DatabaseConnection import DatabaseConnection

logger = logging.getLogger(__name__)


class DatabaseCheck:
    def __init__(self, config):
        self.config = config
        self.dbc = DatabaseConnection.get_connection(self.config)
        assert self.dbc.is_ok()
        self.create_tables()

    def create_tables(self):
        # Create a table to keep track of test items
        self.dbc.query_none(
            """
            create table if not exists "test_access" (
                id serial primary key,
                created_at timestamptz not null default now(),
                updated_at timestamptz not null default now()
            );
            comment on column test_access.id is 'Unique internal id for this item';
            comment on column test_access.created_at is 'When the item was first created';
            comment on column test_access.updated_at is 'When the item was last updated';
            """
        )

    # Insert a new test acceess item into db
    def insert_item(self):
        return self.dbc.query_one(
            """
                insert into test_access default values
                returning id
                ;
                """
        )

    # Get counts of test acceess
    def get_counts(self):
        return self.dbc.query_one(
            """
                select count(*) as count
                from test_access
                ;
                """
        )

    # Get current time from db
    def get_now(self):
        return self.dbc.query_one(
            """
                select now() as now;
                """
        )

    def verify_old(self):
        try:
            count1 = self.get_counts()
            if count1 is None:
                logger.warning("Could not get first count")
                return False
            self.insert_item()
            count2 = self.get_counts()
            if count2 is None:
                logger.warning("Could not get second count")
                return False
            # logger.info(f"count1: {count1}, count2: {count2}")
            return count2.get("count", 0) > count1.get("count", 0)
        except Exception as e:
            logger.warning(f"Unknown error '{e}':", exc_info=True)
            return False

    def verify(self):
        try:
            now = self.get_now()
            return True
        except Exception as e:
            logger.warning(f"Unknown error '{e}':", exc_info=True)
            return False



import os
import pprint

from pathlib import Path
import json
import base64

import ast
import fk.utils
import logging

logger = logging.getLogger(__name__)

# IMPORTANT!!!!!!
# DO NOT PUT REAL CONFIG HERE, ESPECIALLY PASSWORDS AND OTHER SENSITIVE DATA!!!!
unset = "PLEASE OVERRIDE THIS DEFAULT"
defaults = {
    "DEBUG": False,
    "TESTING": False,
    "SECRET_KEY": base64.b64encode(os.urandom(16)).decode("utf-8"),
    "SQLALCHEMY_DATABASE_URI": "sqlite:///:memory:",
    "SQLALCHEMY_ECHO": False,
    "SQLALCHEMY_TRACK_MODIFICATIONS": True,
    "batch-filter-root": "/app/fk/batch/filters",
    "cache-dir-name": "_cache",
    "crawler-html-parser": "html5lib",
    "crawler-referer": "https://www.shopify.com/",
    "crawler-socket-timeout": 15,
    "crawler-user-agent": "Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)",
    "db-database": "postgres",
    "db-hostname": "localhost",
    "db-password": unset,
    "db-port": "5432",
    "db-username": unset,
    "smtp-hostname": unset,
    "smtp-port": "2525",
    "smtp-user": unset,
    "smtp-password": unset,
    "smtp-headers": unset,
    "influx-dbname": "sales",
    "influx-hostname": "localhost",
    "influx-password": unset,
    "influx-port": "8086",
    "influx-username": unset,
    "n2report-dir-name": "_n2report",
    "n2report-index-name": "index.html",
    "overwatch-password": unset,
    "overwatch-user": unset,
    "preferred-url-scheme": "http",
    "printful-api-endpoint": "https://api.printful.com/",
    "printful-api-key": unset,
    "printful-api-key-base64": unset,
    "screenshots-dir-name": "_screenshots",
    "screenshots-index-name": "screenshots.html",
    "shop-id-generator-batch-size": 200,
    "shop-id-generator-max": 25000000000,  # billions
    "shop-id-generator-min": 0,
    "shop-id-thread-count": 300,
    "shopify-api-key": unset,
    "shopify-api-secret": unset,
    "shopify-api-throttle-limit": 1,
    "shopify-api-throttle-period": 3000,
    "shopify-api-version": "2020-10",
    "shopify-app-domain": unset,
    "shopify-app-author": "The author of this app",
    "shopify-app-author-email": "The email of the author of this app",
    "shopify-app-company-name": "The name of the company behind this app",
    "shopify-app-company-website": "The url fo the company website",
    "shopify-app-description": "The description of this app",
    "shopify-app-name": "The name of this app",
    "shopify-app-preferred-url-scheme": "https",
    "shopify-password": unset,
    "shopify-shared-secret": unset,
    "static-files-root": "/app/shopify_app/static",
    "webdriver-max-windows": 5,
    "webdriver-url": "http://127.0.0.1:4444/wd/hub",
    "webdriver-wait-time-sec": 10,
    # "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/40.0.2214.85 Safari/537.36"
    # SHOPIFY_SHOP_NAME = 'firstkissdesigns'
    # https://flask.palletsprojects.com/en/1.1.x/config/#DEBUG
    # https://flask.palletsprojects.com/en/1.1.x/config/#SECRET_KEY
    # SERVER_NAME = "localhost.localdomain:5000"
    # https://flask.palletsprojects.com/en/1.1.x/config/#TESTING
}


class Config(dict):
    def __init__(self):
        super().__init__()
        self.sources = {}

    """Helper to parse boolean from environment variable"""

    def TO_BOOL(str):
        if str.lower() in {"1", "t", "true"}:
            return True
        elif str.lower() in {"0", "f", "false"}:
            return False
        return None

    """Helper to parse unsigned integer from environment variable"""
    # With help from https://stackoverflow.com/questions/1265665/how-can-i-check-if-a-string-represents-an-int-without-using-try-except
    def TO_UINT(str):
        if str.isdigit():
            return int(str)
        if str.startswith("+") and str[1:].isdigit():
            return int(str)
        return None

    """Helper to parse signed integer from environment variable"""

    def TO_INT(str):
        if str.isdigit():
            return int(str)
        if str.startswith(("+", "-")) and str[1:].isdigit():
            return int(str)
        return None

    """Helper to convert a string to environment variable name on the form THIS_IS_ENV_VARIABLE"""

    def TO_ENV_KEY(str):
        str = str.strip().upper()
        for i in ["-", " ", "\t"]:
            str = str.replace(i, "_")
        return str

    """Helper to convert a string to a native python type"""

    def TO_GUESS_TYPE(str):
        b = Config.TO_BOOL(str)
        if b is not None:
            return b
        ret = str
        try:
            ret = ast.literal_eval(str)
        except:
            pass
        return ret

    # @property
    # def DATABASE_URI(self):         # Note: all caps
    #    return 'mysql://user@{}/foo'.format(self.DB_SERVER)

    """Helper to load configuration values from environment variables"""

    def apply_environment_variables(self):
        ct = 0
        for key, value in self.items():
            env_key = Config.TO_ENV_KEY(key)
            env_value = os.environ.get(env_key, default=None)
            if env_value:
                self[key] = Config.TO_GUESS_TYPE(env_value)
                self.sources[key] = "environment"
                ct += 1
                # logger.info(f"Value from env '{env_value}' replaced '{value}' for key '{key}'")
            else:
                # logger.warning(f"WARNING: No {env_key} in env, using last value '{key}' = {value}")
                pass
        logger.info(f"Put {ct} values from environment variables into configuration")
        return True

    def apply_dict(self, other, source="dict"):
        ct = 0
        for key, value in self.items():
            file_value = other.get(key, None)
            if file_value:
                self[key] = file_value
                self.sources[key] = source
                ct += 1
                # logger.info(f"Value from file '{file_value}' replaced '{value}' for key '{key}'")
            else:
                # logger.warning(f"WARNING: No {key} in file, using last value '{value}'")
                pass
        return ct

    """Helper to load configuration values from configuration file in json format"""

    def apply_config_file(self, filename, create_if_missing=True):
        file = Path(filename)
        # Create a config file with defaults if none exists
        if not file.is_file():
            if create_if_missing:
                logger.info(f"Config file '{filename}' did not exist, creating")
                with open(filename, "w") as outfile:
                    json.dump(defaults, outfile, ensure_ascii=False, sort_keys=True, indent=3)
            else:
                logger.info(f"Config file '{filename}' did not exist, skipping")
                logger.info(f"NOTE: current dir was: {os.getcwd()}")
                return False
        # Load config from file
        loaded_config = {}
        if not os.path.exists(filename):
            logger.error(f"ERROR: Could not find '{filename}'")
            return False
        with open(filename, "r") as json_file:
            logger.info(f"Loading config from '{filename}'")
            try:
                loaded_config = json.load(json_file)
                ct = self.apply_dict(loaded_config, filename)
                logger.info(f"Put {ct} values from file variables into configuration")
            except Exception as e:
                logger.error(f"\nParsing json from config file '{filename}' failed:\n{e}\n", exc_info=True)
                return False
        return True

    def list_unchanged(self):
        """ List all variables that have not been changed from their default setting"""
        logger.info(f"Unchanged configurations options:")
        for key, value in defaults.items():
            if self.get(key) == value:
                logger.info(f"{key}={value}")

    def members(self):
        ret = {}
        base = self  # .__class__
        for member in [attr for attr in dir(base) if not callable(getattr(base, attr)) and not attr.startswith("__")]:
            val = getattr(base, member)
            ret[member] = val
        return ret

    def redact_key(self, key, val, num=20, source="redact"):
        if source in ["default", "none"]:
            return val
        if any(x in key.lower() for x in ["secret", "pass", "pwd", "key", "psk"]):
            val = fk.utils.redact(val, num)
        return val

    def __repr__(self):
        dict = self.members()
        ret = f"\n########## Configuration options for {self.__class__.__name__}:\n"
        order = {"none": {}, "default": {}}
        for key, val in dict.items():
            if self.sources != val:
                source = self.sources.get(key, "none")
                order[source] = order.get(source, {})
                order[source][key] = val
        for source, items in order.items():
            if len(items) > 0:
                ret += f"[{source}]\n"
                for key, val in items.items():
                    ret += f" + {key} = {self.redact_key(key, val)}\n"
        return ret

    def redacted_dict(self) -> dict:
        out_dict = dict(self.members())
        for key, val in out_dict.items():
            out_dict[key] = self.redact_key(key, val)
        return out_dict

    def attrify(self):
        for key, val in self.items():
            # logger.debug(f" + {key} = {self.redact_key(key, val)}")
            setattr(self.__class__, key, val)


class DefaultConfig(Config):
    """The default base configuration to be extended by actual configurations to be used"""

    def __init__(self):
        super().__init__()
        self.update(defaults)
        self.sources = {k: "default" for k in self.keys()}

    """Helper to check validity of configuration"""

    def verify(self):
        defkey = []
        for key in ["printful-api-key"]:
            if self.get(key, None) == defaults.get(key, None):
                defkey.append(key)
        if defkey:
            logger.warning(f"Config contains default values for the following keys: '{', '.join(defkey)}'\n        If there are any authentication error this might be the culprit.")

        # Printify needs printful-api-key base64 encoded, so do that now b64encode
        pak = self.get("printful-api-key", None)
        if pak:
            self["printful-api-key-base64"] = str(base64.urlsafe_b64encode(pak.encode("utf-8")).strip(), "utf-8")


"""Configuration that combines defaults with whatever environment variables are set"""


class EnvironmentConfig(DefaultConfig):
    def __init__(self):
        super().__init__()


#        self.apply_environment_variables()


if __name__ == "__main__":
    # Change current working dir to that of this script's location
    cwd = os.path.dirname(os.path.realpath(__file__))
    os.chdir(cwd)
    print("Set working dir to " + cwd)
    print("Config ################################################################")
    c = Config()
    print("---------")
    c.attrify()
    print("---------")
    print(f"CONFIG={c}")
    print("DefaultConfig ################################################################")
    dc = DefaultConfig()
    print("---------")
    dc.verify()
    print("---------")
    dc.attrify()
    print("---------")
    print(f"CONFIG={dc}")
    print("EnvironmentConfig ################################################################")
    ec = EnvironmentConfig()
    print("---------")
    ec.verify()
    print("---------")
    ec.attrify()
    print("---------")
    print(f"CONFIG={ec}")
    print("################################################################")



import sys
import credentials
from Profiler import Profiler


def file_to_dict(filename):
    ret = {}
    with open(filename, "r") as file:
        line = file.readline()
        while line:
            line = line.strip()
            if line != "":
                ret[line] = line
            line = file.readline()
    print(f"Read {len(ret)} lines from {filename}")
    return ret


# Main entrypoint of script
if __name__ == "__main__":

    with Profiler("Elapsed time {}"):
        i = file_to_dict("i_728k.csv")
        o = file_to_dict("o_440k.csv")
        ret = {}
        for il in i:
            if not bool(o.get(il)):
                ret[il] = il

        out_filename = "out_io.csv"
        with open(out_filename, "w") as outfile:
            for ol in ret:
                bob = f"https://gapi.beeketing.com/v1/recsys/recommendation/result/v2/{ol}/best-seller\n"
                # 				outfile.write(f"{ol}\n");
                outfile.write(bob)
            bob = f"https://gapi.beeketing.com/v1/recsys/recommendation/result/v2/{ol}/best-seller\n"
            print(f"Wrote {len(ret)} lines to {out_filename}")



import logging
import sys
import os
import inspect
import warnings

log_setup_once = False


def setup_logging(name=None, log_level=logging.INFO):
    """Set up logging."""
    global log_setup_once
    if not log_setup_once:
        log_setup_once = True
        logging.basicConfig(level=log_level)
        fmt = "%(asctime)s %(levelname)s (%(threadName)s) [%(name)s:%(lineno)s::%(funcName)s()] - %(message)s"
        colorfmt = "%(log_color)s{}%(reset)s".format(fmt)
        datefmt = "%Y-%m-%d %H:%M:%S"

        # Suppress overly verbose output that isn't helpful from some libraries we depend on
        for key in ["requests", "tensorboard", "urllib3", "aiohttp.access", "uamqp", "sqlalchemy", "sqlalchemy.engine.base", "matplotlib.font_manager"]:
            logging.getLogger(key).setLevel(logging.WARNING)

        # Enable debug logging for some insteresting libraries (for development)
        logging.getLogger("fk").setLevel(logging.DEBUG)

        try:
            from colorlog import ColoredFormatter

            logging.getLogger().handlers[0].setFormatter(ColoredFormatter(colorfmt, datefmt=datefmt, reset=True, log_colors={"DEBUG": "cyan", "INFO": "green", "WARNING": "yellow", "ERROR": "red", "CRITICAL": "red"}))
        except ImportError:
            # Ignore failures in loading color coded logs
            pass

        # Set this log level as default
        logger = logging.getLogger("")
        logger.setLevel(log_level)
    if not name:
        name = inspect.stack()[1][1]
    logger = logging.getLogger(name)
    try:
        import colored_traceback.auto
    except ImportError:
        pass

    return logger



from threading import Timer
import logging

logger = logging.getLogger(__name__)


class Watchdog(BaseException):
    def __init__(self, timeout, userHandler=None):  # timeout in seconds
        self.timeout = timeout
        # logger.info(f"WATCHDOG SET WITH TIME {self.timeout}")
        self.handler = userHandler if userHandler is not None else self.defaultHandler
        self.timer = Timer(self.timeout, self.handler)
        self.timer.start()

    def reset(self):
        self.timer.cancel()
        self.timer = Timer(self.timeout, self.handler)
        self.timer.start()

    def stop(self):
        # logger.info(f"WATCHDOG STOPPED")
        self.timer.cancel()

    def defaultHandler(self):
        logger.info("WATCHDOG TIMED OUT")
        raise self



from pathlib import Path
import json
import sys
import os
import base64
from .utils import merge

# IMPORTANT!!!!!!
# DO NOT PUT REAL CREDENTIALS HERE!
# INSTEAD EDIT credentials.json THAT IS
# AUTOMATICALLY CREATED ON FIRST RUN OF MAIN SCRIPT
default_credentials = {
    "printful-api-key-source": "Get it here: https://firstkissdesign.myshopify.com/admin/apps/3dd87416a3b1e12a9ef3604aaaaf4d1e/dashboard/settings/1526227/store-api",
    "printful-api-key": "REPLACE-WITH-ACTUAL-KEY",
    "printful-api-key-base64": "",
    "printful-api-endpoint": "https://api.printful.com/",
    "shopify-api-key": "",
    "shopify-password": "",
    "shopify-api-version": "",
    "shopify-shared-secret": "",
    "shopify-shop-name": "",
    "webdriver-url": "http://127.0.0.1:4444/wd/hub",
    "webdriver-max-windows": 5,
    "webdriver-wait-time-sec": 10,
    "influx-hostname": "localhost",
    "influx-port": "8086",
    "influx-username": "root",
    "influx-password": "root",
    "influx-dbname": "sales",
    "cache-dir-name": "_cache",
    "n2report-dir-name": "_n2report",
    "n2report-index-name": "index.html",
    "screenshots-dir-name": "_screenshots",
    "screenshots-index-name": "screenshots.html",
    "db-hostname": "localhost",
    "db-port": "5432",
    "db-username": "postgres",
    "db-password": "",
    "db-database": "postgres",
    "shop-id-generator-min": 0,
    "shop-id-generator-max": 25000000000,  # billions
    "shop-id-generator-batch-size": 200,
    "shop-id-thread-count": 300,
    "crawler-socket-timeout": 15,
    "crawler-referer": "https://www.shopify.com/",
    "crawler-user-agent": "Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)",
    # "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/40.0.2214.85 Safari/537.36"
    "crawler-html-parser": "html5lib",
}


# Load credentials from file, making sure to create defaults if no credentials file found
def load_credentials_ba(credentials_fn="credentials.json"):
    credentials_file = Path(credentials_fn)
    # Create a credentials file with defaults if none exists
    if not credentials_file.is_file():
        print(f"Credentials file {credentials_fn} did not exist, creating")
        with open(credentials_fn, "w") as outfile:
            json.dump(default_credentials, outfile, ensure_ascii=False, sort_keys=True, indent=3)
    # Load credentials from file
    credentials = {}
    if not os.path.exists(credentials_fn):
        print(f"ERROR: Could not find {credentials_fn}")
        sys.exit(1)
    with open(credentials_fn, "r") as json_file:
        print(f"Loading credentials from {credentials_fn}")
        try:
            loaded_credentials = json.load(json_file)
        except Exception as e:
            print("")
            print(f"ERROR: parsing json from '{credentials_fn}':")
            print("")
            print(f"	   {e}")
            print("")
            return None
        credentials = merge(default_credentials, loaded_credentials)

    if credentials["printful-api-key"] == default_credentials["printful-api-key"]:
        print(f"NOTICE: {credentials_fn} contains default values for some credentials\n		If there are any authentication error this might be the culprit.")

    # Printify needs printful-api-key base64 encoded, so do that now b64encode
    credentials["printful-api-key-base64"] = str(base64.urlsafe_b64encode(credentials["printful-api-key"].encode("utf-8")).strip(), "utf-8")

    # Print credentials to terminal for reference
    # print('Credentials: ') 	pprint.pprint(credentials)
    return credentials



from dateutil.relativedelta import relativedelta
from urllib.parse import urlparse
import hashlib
import os
import json
import mmap
import base64
import random
import datetime
import time
import requests
import pkg_resources
from uuid import UUID
import logging

logger = logging.getLogger(__name__)

# We want 4 second timeout for HTTP requests
HTTP_TIMEOUT = 4.0
_requests_session_instance = None


def redact(val, minimum=20):
    total = len(val)
    scramble_n = max(minimum, total - 3)
    show_n = max(min(total - minimum, 3), 0)
    out = val[0:show_n] + "*" * (scramble_n)
    return out


def file_contains_str(filename, str):
    return file_contains_bytes(filename, str.encode("utf-8"))


def file_contains_bytes(filename, bytes):
    # mmap does not work for empty files
    if os.stat(filename).st_size == 0:
        return False
    with open(filename, "rb", 0) as file, mmap.mmap(file.fileno(), 0, access=mmap.ACCESS_READ) as s:
        if s.find(bytes) != -1:
            return True
    return False


# Merge two dictionaries
def merge(dict1, dict2):
    res = {**dict1, **dict2}
    return res


# Determine if given URL is valid
def is_url_valid(url):
    try:
        result = urlparse(url)
        # 		if not result.scheme:
        # 			print ("No scheme")
        # 		if not result.netloc:
        # 			print ("No netloc")
        # 		if not result.path:
        # 			print ("No path")
        # 		if not result.scheme in ['http', 'https']:
        # 			print ("Scheme invalid")

        return all([result.scheme, result.netloc]) and result.scheme in ["http", "https"]
        # result.netloc,
    except:
        return False


def is_valid_uuid(uuid_to_test, version=4):
    """
    Check if uuid_to_test is a valid UUID.

    Parameters
    ----------
    uuid_to_test : str
    version : {1, 2, 3, 4}

    Returns
    -------
    `True` if uuid_to_test is a valid UUID, otherwise `False`.

    Examples
    --------
    >>> is_valid_uuid('c9bf9e57-1685-4c89-bafb-ff5af830be8a')
    True
    >>> is_valid_uuid('c9bf9e58')
    False
    """
    try:
        uuid_obj = UUID(uuid_to_test, version=version)
    except ValueError:
        return False

    return str(uuid_obj) == uuid_to_test


# Create a hash hex digest from string
def hashstr(raw):
    if type(raw) != type(""):
        raw = f"{raw}"
    sha256 = hashlib.sha256()
    sha256.update(raw.encode("utf-8"))
    return sha256.hexdigest().upper()


def hashfile(fn):
    file = open(fn, "r+")
    size = os.path.getsize(fn)
    sha256 = hashlib.sha256()
    sha256.update(mmap.mmap(file.fileno(), size))
    return sha256.update(mmap.mmap(file.fileno(), size))


# Turn invalid string into naive URL by prepending schema and appending path
def decorate_url(url):
    if not is_url_valid(url):
        url = "http://{url}/".format(url=url)
    return url


# Return dict with url split into components as appropriate for storing in database
def split_url(url_full):
    ret = {"url_hostname": "", "url_schema": "", "url_path": "", "url_query": "", "url_fragment": ""}
    try:
        parts = urlparse(url_full)
        if parts.netloc is not None:
            ret["url_hostname"] = parts.netloc
        if parts.scheme is not None:
            ret["url_schema"] = parts.scheme
        if parts.path is not None:
            ret["url_path"] = parts.path
        if parts.query is not None:
            ret["url_query"] = parts.query
        if parts.port is not None:
            ret["url_port"] = parts.port
        if parts.fragment is not None:
            ret["url_fragment"] = parts.fragment
    except Exception as e:
        # print(f"ERROR SPLITTING URL: {e}")
        pass
    return ret


def flatten_headers(headers):
    out = ""
    for k, v in headers.items():
        out += f"'{k}'='{v}'\n"
    return out


def verify_input_files(input_options):
    input_files = []
    notes = []
    if input_options:
        for file in input_options:
            note = "[OK]"
            if not os.path.exists(file):
                note = "[MISSING, Skipped]"
            elif not os.path.isfile(file):
                note = "[NOT A FILE, Skipped]"
            elif os.access(file, os.R_OK):
                note = "[NOT READABLE, Skipped]"
            else:
                input_files.append(file)
            notes.append(f' "{file}"   {note}')
        return input_files, notes
    else:
        return None, None


def verify_output_file(output_file):
    valid_output_file = None
    if output_file:
        output_path = os.path.dirname(os.path.realpath(output_file))
        os.makedirs(output_path, exist_ok=True)
        note = "[OK]"
        if not os.path.exists(output_path):
            note = "[PATH MISSING]"
        elif os.access(output_file, os.W_OK):
            note = "[NOT WRITABLE]"
        else:
            valid_output_file = output_file
    return valid_output_file, note


# Look at result from scrape and produce a dictionary with relevant data or error output
def get_data_for_scrape_result(result):
    http_status_code = result.get("http_status_code", 500)
    reason = result.get("reason", "")
    print(f"SCRAPING COMPLETED {http_status_code}, {reason}")

    data = None
    error = ""
    ok = True
    if http_status_code != 200:
        error = f"http code {http_status_code} with reason {reason}"
        ok = False
    else:
        headers = result.get("headers", "")
        page_source = result.get("page_source", "")
        if not page_source:
            error = "no page_source"
            ok = False
        else:
            data = None
            try:
                data = json.loads(page_source)
            except Exception as e:
                error = f"json parse failed for {page_source[:100]} with {e}"
                ok = False
            if ok:
                t = __builtins__.type(data)
                if not data:
                    error = f"json parse returned no data"
                    ok = False
                # elif t is not '<class \'dict\'>':
                # 	error=f"json parse returned not dict ({t})"
                # 	ok=False
    return data, ok, error


attrs = ["millenia", "centuries", "decades", "years", "months", "days", "hours", "minutes", "seconds"]

human_readable = lambda delta: ["%d %s" % (getattr(delta, attr), getattr(delta, attr) > 1 and attr or attr[:-1]) for attr in attrs if getattr(delta, attr)]


def human_delta(td_object: datetime.timedelta, max: int = 0):
    ms = int(td_object.total_seconds() * 1000)
    if ms == 0:
        return "0 ms"
    sign = ""
    if ms < 0:
        ms = -ms
        sign = "-"
    # fmt: off
    periods = [
        ("year",  1000 * 60 * 60 * 24 * 365),
        ("month", 1000 * 60 * 60 * 24 * 30),
        ("day",   1000 * 60 * 60 * 24),
        ("hr",    1000 * 60 * 60),
        ("min",   1000 * 60),
        ("sec",   1000),
        ("ms", 1)
    ]
    # fmt: on

    strings = []
    ct: int = 0
    for period_name, period_ms in periods:
        if ms > period_ms:
            period_value, ms = divmod(ms, period_ms)
            # has_s = "s" if period_value > 1 else ""
            # strings.append("%s %s%s" % (period_value, period_name, has_s))
            strings.append(f"{period_value} {period_name}")
            ct += 1
            if max > 0 and ct > max:
                break
    return sign + ", ".join(strings)  # + f"({td_object}, {ms})"


def human_bytesize(bytes: int, max: int = 0):
    if bytes == 0:
        return "0 bytes"
    if bytes < 0:
        return "negative bytes (error)"
    # fmt: off
    KiB=1024
    periods = [
        ("PiB", KiB*KiB*KiB*KiB*KiB),
        ("TiB", KiB*KiB*KiB*KiB),
        ("GiB", KiB*KiB*KiB),
        ("MiB", KiB*KiB),
        ("KiB", KiB),
        ("bytes", 1)
    ]
    # fmt: on

    strings = []
    ct: int = 0
    for period_name, period_bytes in periods:
        if bytes > period_bytes:
            period_value, bytes = divmod(bytes, period_bytes)
            # has_s = "s" if period_value > 1 else ""
            # strings.append("%s %s%s" % (period_value, period_name, has_s))
            strings.append(f"{period_value} {period_name}")
            ct += 1
            if max > 0 and ct > max:
                break
    return ", ".join(strings)  # + f"({td_object}, {bytes})"


# Make washee conform strictly to structure of washer
def wash_dict(washee, washer):
    out = {}
    for k, v in washer.items():
        if k in washee:
            if isinstance(v, dict):
                if isinstance(washee[k], dict):
                    out[k] = wash_dict(washee[k], v)
            else:
                if not isinstance(washee[k], dict):
                    out[k] = washee[k]
    return out


def random_str(l):
    return base64.b64encode(os.urandom(l))


def print_process_info():
    if hasattr(os, "getppid"):
        print(f"Parent process:{os.getppid()}")
    print(f"Process id:{os.getpid()}")


def sleep(time_sec):
    time.sleep(time_sec)
    # await asyncio.sleep async


def read_file(fname, strip=True):
    fn = os.path.join(os.path.dirname(os.path.abspath(__file__)), fname)
    data = ""
    if os.path.exists(fn):
        with open(fn, encoding="utf-8") as f:
            data = f.read()
            data = data.strip() if strip else data
            # logger.info(f"Got data '{data}' from '{fn}'")
    else:
        logger.error(f"Could not find file {fn} relative to working directory {os.getcwd()}")
    return data


def write_file(fn, data, strip=True):
    with open(fn, "w") as f:
        f.write(data.strip() if strip else data)


def debug_imports():
    import logging
    import pprint
    import sys
    import os

    l = logging.getLogger(__name__)
    l.error(f"PATH: {sys.path}")
    l.error(f"PREFIX: {sys.prefix}")
    l.error(f"ENV: {pprint.pformat(os.environ)}")

    import pkg_resources

    installed_packages = {d.project_name: d.version for d in pkg_resources.working_set}
    l.error(f"MODULES:{pprint.pformat(installed_packages)}")

    import pkgutil

    for m in pkgutil.iter_modules(path=None):
        l.error(f"MODULE: {m.name}")

    def explore_package(module_name):
        loader = pkgutil.get_loader(module_name)
        if not loader:
            l.error(f"No loader for {module_name}")
        elif not loader.filename:
            l.error(f"No filename for {module_name}")
        else:
            for sub_module in pkgutil.walk_packages([loader.filename]):
                _, sub_module_name, _ = sub_module
                qname = module_name + "." + sub_module_name
                l.error(qname)
                explore_package(qname)

    explore_package("fk")
    explore_package("batch")

    import time

    time.sleep(die_sec)
    sys.exit(6)


class TimeoutRequestsSession(requests.Session):
    def request(self, *args, **kwargs):
        kwargs.setdefault("timeout", HTTP_TIMEOUT)
        return super(TimeoutRequestsSession, self).request(*args, **kwargs)


def requests_session():
    global _requests_session_instance
    if not _requests_session_instance:
        _requests_session_instance = TimeoutRequestsSession()
        # fmt:off
        headers = {
              "User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/40.0.2214.85 Safari/537.36"
            , "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9"
            , "Cache-Control": "no-cache"
            , "Pragma": "no-cache"
        }
        # fmt:on
        _requests_session_instance.headers.update(headers)
    return _requests_session_instance


#import os
#import pkg_resources
#from fk.utils import read_file
def get_package_file_content_(filename, do_debug=True):
    """Look for file relative to package root and return it's content"""
    content = ""
    package_base_path: str = os.path.abspath(os.path.join(os.path.dirname(__file__), "../"))
    script_path: str = os.path.abspath(os.path.dirname(__file__))
    full_filename = os.path.join(package_base_path, filename)
    if do_debug:
        logger.warn(f"package_base_path={package_base_path}")
        logger.warn(f"script_path={script_path}")
        logger.warn(f"filename={filename}")
        logger.warn(f"__name__={__name__}")
        logger.warn(f"__file__={__file__}")
        logger.warn(f"full_filename={full_filename}")
    if pkg_resources.resource_exists(__name__, filename):
        logger.info(f"Fetching {filename} from package")
        content = pkg_resources.resource_string(__name__, filename).decode("utf-8").strip()
    elif os.path.exists(full_filename):
        logger.info(f"Fetching {full_filename} from dir")
        content = read_file(full_filename)
    else:
        logger.warning(f"No content found for {full_filename}")
    return content


### ^ REMIX VERSION

def get_package_file_content(filename, do_debug=True):
    """Look for file relative to package root and return it's content"""
    content = ""
    package_base_path: str = os.path.abspath(os.path.join(os.path.dirname(__file__), "../"))
    script_path: str = os.path.abspath(os.path.dirname(__file__))
    full_filename = os.path.join(package_base_path, filename)
    if do_debug:
        logger.warn(f"package_base_path={package_base_path}")
        logger.warn(f"script_path={script_path}")
        logger.warn(f"filename={filename}")
        logger.warn(f"__name__={__name__}")
        logger.warn(f"__file__={__file__}")
        logger.warn(f"full_filename={full_filename}")
    if pkg_resources.resource_exists(__name__, filename):
        logger.info(f"Fetching {filename} from package")
        content = pkg_resources.resource_string(__name__, filename).decode("utf-8").strip()
    elif os.path.exists(full_filename):
        logger.info(f"Fetching {full_filename} from dir")
        content = read_file(full_filename)
    else:
        logger.warning(f"No content found for {full_filename}")
    return content



from time import time
import datetime
import math


def _human_time(*args, **kwargs):
    secs = float(datetime.timedelta(*args, **kwargs).total_seconds())
    units = [("day", 86400), ("hour", 3600), ("minute", 60), ("second", 1)]
    parts = []
    for unit, mul in units:
        if secs / mul >= 1 or mul == 1:
            if mul > 1:
                n = int(math.floor(secs / mul))
                secs -= n * mul
            else:
                n = secs if secs != int(secs) else int(secs)
            parts.append("%s %s%s" % (n, unit, "" if n == 1 else "s"))
    return ", ".join(parts)


class Profiler:
    def __init__(self, message):
        self.message = message

    def __enter__(self):
        self.start = time()
        return None  # could return anything, to be used like this: with Timer("Message") as value:

    def __exit__(self, type, value, traceback):
        sec = time() - self.start
        elapsed_time = _human_time(seconds=sec)
        print(f"{self.message} took {format(elapsed_time)}")



from influxdb import InfluxDBClient
from pprint import pformat
import traceback
import logging

logger = logging.getLogger(__name__)


# Helper class to manage influxdb
class Influx:
    def __init__(self, credentials):
        logger = setup_logging(self.__class__.__name__)
        self.credentials = credentials
        self.client = InfluxDBClient(self.credentials["influx-hostname"], self.credentials["influx-port"], self.credentials["influx-username"], self.credentials["influx-password"], self.credentials["influx-dbname"])
        self.client.create_database(self.credentials["influx-dbname"])
        self.transaction_points = None
        self.transaction_tags = {}
        self.transaction_time_precision = "ms"

    # Begin super basic transaction
    def begin_transaction(self, time_precision="ms"):
        if None is not self.transaction_points:
            logger.error("Another transaction is in progress, aborting")
            return False
        self.transaction_points = []
        self.transaction_tags = {}
        self.transaction_time_precision = time_precision
        return True

    # Add point to super basic transaction
    def add_point_transaction(self, measurement, time, tags={}, fields={}):
        if None is self.transaction_points:
            logger.error("Transaction not started, aborting")
            return False
        # Keep tags in set
        for key, val in tags.items():
            self.transaction_tags[key] = val
        # Keep point
        point_body = {"measurement": measurement, "tags": tags, "time": time, "fields": fields}
        self.transaction_points.append(point_body)
        return True

    # Commit super basic transaction
    def commit_transaction(self):
        ret = False
        if self.transaction_points is not None:
            try:
                ret = self.client.write_points(points=self.transaction_points, time_precision=self.transaction_time_precision, tags=self.transaction_tags)
            except Exception as e:
                logger.error(f" ")
                logger.error(f"#########################################################")
                logger.error(f"## ERROR: Could not commit influx transaction")
                logger.error(f"## Transaction details':")
                logger.error(f"##")
                logger.error(f"## time_precision={self.transaction_time_precision}")
                logger.error(f"## points=")
                logger.error(pformat(self.transaction_points))
                logger.error(f"## tags=")
                logger.error(pformat(self.transaction_tags))
                logger.error(f"##")
                logger.error(f"## {e}:")
                logger.error(f"##")
                traceback.print_exc()
                logger.error(f"##")
                logger.error(f"#########################################################")
                logger.error(f" ")
            self.transaction_points = None
            self.transaction_tags = {}
        return ret

    # Simple single point insertion
    def put_point(self, measurement, time, tags={}, fields={}):
        body = [{"measurement": measurement, "tags": tags, "time": time, "fields": fields}]
        return self.client.write_points(body)

    # Perform query
    def query(self, q):
        return self.client.query(q)



import json
import os
import sqlite3
import pprint
import psycopg2
import psycopg2.extras
import traceback
import sys
import logging
from urllib.parse import urlparse, urlunparse, ParseResult
import pytz
import fk.utils
import unicodedata
from furl import furl
import pprint

logger = logging.getLogger(__name__)


class DatabaseConnection:

    # Global connection
    connection_instances = {}

    def __init__(self, config):
        self.config = config
        self.db_hostname = self.config.get("db-hostname", None)
        self.db_port = self.config.get("db-port", None)
        self.db_username = self.config.get("db-username", None)
        self.db_password = self.config.get("db-password", None)
        self.db_database = self.config.get("db-database", None)
        self.db = None
        connstr = DatabaseConnection.config_to_url(config, do_redact=True)
        # connstr = f"postgres://{self.db_username}:{fk.utils.redact(self.db_password)}@{self.db_hostname}:{self.db_port}/{self.db_database}"
        logger.info(f"DATABASE: {connstr}")
        self.connection_error = None
        self._prepare_db()

    def __del____(self):
        self._unprepare_db()

    def is_ok(self):
        return self.db_hostname and self.db_port and self.db_username and self.db_password and self.db_database

    @staticmethod
    def config_to_url(config, do_redact=False):
        if not config:
            return ""
        db_hostname = config.get("db-hostname", None)
        db_port = config.get("db-port", None)
        db_username = config.get("db-username", None)
        db_password = config.get("db-password", None)
        db_database = config.get("db-database", None)
        if do_redact:
            db_password = fk.utils.redact(db_password)
        f = furl()
        f.set(scheme="postgres", username=db_username, password=db_password, host=db_hostname, port=db_port, path=db_database)
        return f.url

    @staticmethod
    def url_to_config(db_url):
        f = furl(db_url)
        if f.scheme != "postgres":
            logger.warn("db_url URL did not have postgres scheme")
            return {}
        database = (f.path.segments or [""])[0]
        # fmt:off
        config={
          "db-hostname": f.host
        , "db-port": f.port
        , "db-username": f.username
        , "db-password": f.password
        , "db-database": database
        }
        # fmt:on
        return config

    @staticmethod
    def get_config_hash(config):
        if not config:
            return ""
        key = ""
        for i in ["hostname", "port", "username", "password", "database"]:
            key += fk.utils.hashstr(config.get(f"db-{i}", ""))
        return fk.utils.hashstr(key)

    @staticmethod
    def get_connection(config):
        key = DatabaseConnection.get_config_hash(config)
        if not key in DatabaseConnection.connection_instances:
            DatabaseConnection.connection_instances[key] = DatabaseConnection(config)
        return DatabaseConnection.connection_instances[key]

    def _unprepare_db(self):
        if self.db:
            try:
                # Make sure data is stored
                self.db.commit()
                self.db.close()
            except Exception as e1:
                pass
            try:
                del self.db
            except Exception as e2:
                pass
            self.db = None
            logger.info("PostgreSQL connection is closed")

    # Internal helper to connect to database
    def _prepare_db(self):
        try:
            self.db = psycopg2.connect(host=self.db_hostname, port=self.db_port, user=self.db_username, password=self.db_password, database=self.db_database)

            # Create a cursor to let us work with the database
            with self.db:
                with self.db.cursor() as c:
                    # logger.info( self.db.get_dsn_parameters(),"\n")

                    c.execute("SELECT version();")
                    record = c.fetchone()
                    logger.info(f"Connected to: {record[0]}\n")
                    return True

        except (Exception, psycopg2.Error) as error:
            logger.warning("###   ###   ###   ###   ###   ###   ###   #")
            logger.warning("##   ###   ###   ###   ###   ###   ###   ##")
            logger.warning("")
            logger.warning(f"Error while connecting to PostgreSQL {error}")
            self.connection_error = error
            self._unprepare_db()
        return False

    def reconnect(self):
        logger.info("###   ###   ###   ###   ###   ###   ###   #")
        logger.info("##   ###   ###   ###   ###   ###   ###   ##")
        logger.info("")
        logger.info(f"Reconnecting...")
        self._unprepare_db()
        return self._prepare_db()

    def _query(self, query, data={}, mode="none"):
        cursor = None
        try_reconnect = True
        try:
            if not self.db:
                logger.warning("###   ###   ###   ###   ###   ###   ###   #")
                logger.warning("##   ###   ###   ###   ###   ###   ###   ##")
                logger.warning("")
                logger.warning("No database while making query")
                if self.connection_error:
                    logger.warning(f"Connection error was {self.connection_error}")
                else:
                    logger.warning("No connection error was set, did we even try to connect?")
            with self.db:
                with self.db.cursor(cursor_factory=psycopg2.extras.DictCursor) as cursor:
                    cursor.execute(query, data)
                    if mode == "many":
                        res = cursor.fetchall()
                        # Convert to real dictionaries
                        if res is None:
                            return None
                        out = []
                        for row in res:
                            out.append(dict(row))
                        return out
                    elif mode == "one":
                        res = cursor.fetchone()
                        if res is None:
                            return None
                        return dict(res)
                    elif mode == "exists":
                        res = cursor.fetchone()
                        return res is not None
                    else:
                        return
        except Exception as e:
            logger.error("")
            logger.error("")
            logger.error(f"###############################################")
            logger.error(f"#    Querying: {query}")
            if cursor and cursor.query:
                logger.error(f"#    Injected: {cursor.query.decode()}")
            logger.error(f"#         For: {mode}")
            logger.error(f"# Failed with:")
            if isinstance(e, psycopg2.InterfaceError):
                logger.error(f"# + Connection error({type(e).__name__}): e={e} ")
                try_reconnect = True
            elif isinstance(e, psycopg2.Error):
                logger.error(f"# + e={e} ({type(e).__name__})")
                logger.error(f"# + e.pgerror:{e.pgerror}")
                logger.error(f"# + e.pgcode: {e.pgcode}")
                logger.error(f"# + e.cursor: {e.cursor}")
                if e.diag:
                    logger.error(f"# + e.diag:")
                    logger.error(f"#   + column_name:             {e.diag.column_name}")
                    logger.error(f"#   + constraint_name:         {e.diag.constraint_name}")
                    logger.error(f"#   + context:                 {e.diag.context}")
                    logger.error(f"#   + datatype_name:           {e.diag.datatype_name}")
                    logger.error(f"#   + internal_position:       {e.diag.internal_position}")
                    logger.error(f"#   + internal_query:          {e.diag.internal_query}")
                    logger.error(f"#   + message_detail:          {e.diag.message_detail}")
                    logger.error(f"#   + message_hint:            {e.diag.message_hint}")
                    logger.error(f"#   + message_primary:         {e.diag.message_primary}")
                    logger.error(f"#   + schema_name:             {e.diag.schema_name}")
                    logger.error(f"#   + severity:                {e.diag.severity}")
                    logger.error(f"#   + source_file:             {e.diag.source_file}")
                    logger.error(f"#   + source_function:         {e.diag.source_function}")
                    logger.error(f"#   + source_line:             {e.diag.source_line}")
                    logger.error(f"#   + sqlstate:                {e.diag.sqlstate}")
                    logger.error(f"#   + statement_position:      {e.diag.statement_position}")
                    logger.error(f"#   + table_name:              {e.diag.table_name}")
                    logger.error(f"#   + severity_nonlocalized:   {e.diag.severity_nonlocalized}")
            else:
                logger.error(f"# + NON DB ERROR({type(e).__name__}): e={e} ")
            logger.error(f"#          At:")
            traceback.print_exception(type(e), e, e.__traceback__)
            logger.error(f"#       Stack:")
            traceback.print_stack()
            logger.error(f"#       Using data:")
            logger.error(pprint.pformat(data))
            logger.error(f"#######################################")
            logger.error("")
            logger.error("")
        if try_reconnect:
            self.reconnect()

    def query_many(self, query, data={}):
        return self._query(query, data, "many")

    def query_one(self, query, data={}):
        return self._query(query, data, "one")

    def query_none(self, query, data={}):
        return self._query(query, data, "none")

    def query_exists(self, query, data={}):
        return self._query(query, data, "exists")



import datetime
import math
from time import sleep

from fk.utils import human_delta

import logging

logger = logging.getLogger(__name__)


class OnTheClockTimer:
    """
    Class to provide a schedule from time of day + recurring interval as starting point
    Also provides convenience methods to make it easy to keep track of this schedule
    """

    def __init__(self, start_time: datetime.time, interval: datetime.timedelta):
        self.start_time = start_time
        self.interval = interval

    def closest_start_time(self, now: datetime.datetime = datetime.datetime.now()) -> datetime.datetime:
        time_of_day = now.time()
        common_date = datetime.date(2019, 1, 1)
        p1 = datetime.datetime.combine(common_date, time_of_day)
        p2 = datetime.datetime.combine(common_date, self.start_time)
        from_start = p1 - p2
        interval_count = from_start / self.interval
        next_time = (math.floor(interval_count) + 1) * self.interval
        # logger.info(f"DATE: {now.date()}, TIME:{now.time()}, from_start:{from_start}, interval_count:{interval_count}, last_time={last_time}, next_time={next_time}")
        return datetime.datetime.combine(now.date(), self.start_time) + next_time

    def time_left(self, now: datetime.datetime = datetime.datetime.now()) -> datetime.timedelta:
        return self.closest_start_time(now=now) - now

    def wait_for_trigger(self, now: datetime.datetime = datetime.datetime.now()):
        iv = self.time_left(now=now)
        sec = iv.total_seconds()
        if sec > 0:
            logger.info(f"Waiting for {human_delta(iv)}")
            sleep(sec)
        return True

    def __str__(self):
        return f"OnTheClockTimer(start_time={self.start_time}, interval={self.interval})"



from Database import Database
from Influx import Influx
from queue import Queue
from ScraperFactory import ScraperFactory
import credentials
import json
import os
import pprint
import re
import sys
import time
import utils
import base64
import csv
import datetime


# def unicode_csv_reader(utf8_data, dialect=csv.excel, **kwargs):
# 	csv_reader = csv.reader(utf8_data, dialect=dialect, **kwargs)
# 	for row in csv_reader:
# 		yield [unicode(cell, 'ISO-8859-1') for cell in row]


# Main entrypoint of script
if __name__ == "__main__":
    # Change current working dir to that of this script's location
    cwd = os.path.dirname(os.path.realpath(__file__))
    os.chdir(cwd)
    print("Set working dir to " + cwd)

    # Load credentials
    credentials = credentials.load_credentials()
    if not credentials:
        sys.exit(1)

    print("#############################################################################")

    db = Database(credentials)

    # File containting one shop name per line
    oh_no_fn = "oh_no.csv"

    # reader = unicode_csv_reader(open(oh_no_fn))

    # for site_id in reader:		print(site_id)

    map = {"Comment": "comment", "Api Key": "api_key", "Created At": "created_at", "Fetched At": "fetched_at", "Fetched Count": "fetched_count", "Link": "link", "Sale Date First": "sale_date_first", "Sale Date First Ever": "sale_date_first_ever", "Sale Date Last": "sale_date_last", "Site Id": "site_id", "Spread": "spread"}

    # sys.exit(1)
    with open(oh_no_fn) as csv_file:
        csv_reader = csv.reader(csv_file, delimiter="\t")
        line_count = 0
        headers = []
        for row in csv_reader:
            if line_count == 10:
                # break
                pass
            if line_count == 0:
                print(f'Column names are {", ".join(row)}')
                headers = row
                line_count += 1
            else:
                spop = {}
                for index, header in enumerate(headers):
                    h = map.get(header, header)
                    v = row[index]
                    if v == "":
                        v = None
                    elif h.startswith("sale_date_"):
                        v = datetime.datetime(v)
                    spop[h] = v
                i = db.upsert_spop_id(spop)
                # print(f"LINE:{spop} ---> {i}")
                line_count += 1
        print(f"Processed {line_count} lines.")



from __future__ import print_function
from queue import Queue
from threading import Thread, Event, Lock
from sys import stdout, stderr
from time import sleep
import sys


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


# default exception handler. if you want to take some action on failed tasks
# maybe add the task back into the queue, then make your own handler and pass it in
def default_handler(name, exception, *args, **kwargs):
    eprint(f"{name} raised {exception} with args {repr(args)} and kwargs {repr(kwargs)}")
    pass


# class for workers
class Worker(Thread):

    """Thread executing tasks from a given tasks queue"""

    def __init__(self, name, queue, results, abort, idle, exception_handler):
        Thread.__init__(self)
        self.name = name
        self.queue = queue
        self.results = results
        self.abort = abort
        self.idle = idle
        self.exception_handler = exception_handler
        self.daemon = True
        self.start()

    """Thread work loop calling the function with the params"""

    def run(self):
        # keep running until told to abort
        while not self.abort.is_set():
            try:
                # get a task and raise immediately if none available
                func, args, kwargs = self.queue.get(False)
                self.idle.clear()
            except:
                # no work to do
                # if not self.idle.is_set():
                #  print >> stdout, '%s is idle' % self.name
                self.idle.set()
                continue

            try:
                # the function may raise
                result = func(*args, **kwargs)
                if result is not None:
                    self.results.put(result)
            except Exception as e:
                # so we move on and handle it in whatever way the caller wanted
                self.exception_handler(self.name, e, args, kwargs)
            finally:
                # task complete no matter what happened
                self.queue.task_done()


# class for thread pool
class Pool:

    """Pool of threads consuming tasks from a queue"""

    def __init__(self, thread_count, batch_mode=False, exception_handler=default_handler):
        # batch mode means block when adding tasks if no threads available to process
        self.queue = Queue(thread_count if batch_mode else 0)
        self.resultQueue = Queue(0)
        self.thread_count = thread_count
        self.exception_handler = exception_handler
        self.aborts = []
        self.idles = []
        self.threads = []

    """Tell my threads to quit"""

    def __del__(self):
        self.abort()

    """Start the threads, or restart them if you've aborted"""

    def run(self, block=False):
        # Either wait for them to finish or return false if some aren't
        if block:
            while self.alive():
                sleep(1)
        elif self.alive():
            return False

        # Start them
        self.aborts = []
        self.idles = []
        self.threads = []
        for n in range(self.thread_count):
            abort = Event()
            idle = Event()
            self.aborts.append(abort)
            self.idles.append(idle)
            self.threads.append(Worker("thread-%d" % n, self.queue, self.resultQueue, abort, idle, self.exception_handler))
        return True

    """Add a task to the queue"""

    def enqueue(self, func, *args, **kargs):
        self.queue.put((func, args, kargs))

    """Wait for completion of all the tasks in the queue"""

    def join(self):
        self.queue.join()

    """Tell each worker that its done working"""

    def abort(self, block=False):
        # tell the threads to stop after they are done with what they are currently doing
        for a in self.aborts:
            a.set()
        # wait for them to finish if requested
        while block and self.alive():
            sleep(1)

    """Returns True if any threads are currently running"""

    def alive(self):
        return True in [t.is_alive() for t in self.threads]

    """Returns True if all threads are waiting for work"""

    def idle(self):
        return False not in [i.is_set() for i in self.idles]

    """Returns True if not tasks are left to be completed"""

    def done(self):
        return self.queue.empty()

    """Get the set of results that have been processed, repeatedly call until done"""

    def results(self, wait=0):
        sleep(wait)
        results = []
        try:
            while True:
                # get a result, raises empty exception immediately if none available
                results.append(self.resultQueue.get(False))
                self.resultQueue.task_done()
        except:
            pass
        return results


# run the test
if __name__ == "__main__":

    def work(x):
        sleep(5)
        print >> stdout, "%d finished" % x
        return x

    def wait_for_results(p):
        while not p.done() or not p.idle():
            for result in p.results():
                print >> stdout, "got result %s" % str(result)
        for result in p.results():
            print >> stdout, "got result %s" % str(result)
        p = Pool(3)
        print >> stdout, "queueing work"
        for i in range(10):
            p.enqueue(work, i)
        print >> stdout, "starting work"
        p.run(True)
        print >> stdout, "waiting"
        sleep(8)
        for result in p.results():
            print >> stdout, "got result %s" % str(result)
        print >> stdout, "cancel unstarted work"
        p.abort()
        print >> stdout, "waiting and restarting"
        p.run(True)
        print >> stdout, "restarted waiting for results"
        wait_for_results(p)
        print >> stdout, "adding more work"
        for i in range(10, 17):
            p.enqueue(work, i)
        wait_for_results(p)



import pprint
from flask import Blueprint, render_template, make_response, current_app, request, redirect, session, url_for, send_from_directory, jsonify, abort, g

# return debugger(request, session, "Install page", permission_url)
# return debugger(request, session, pprint.pformat(shop), redirect_url)
# return debugger(request, session, "Main page", redirect_url)


def debugger(request, session, message, link):
    return render_template("shopify_bp/debugger.html", request=pprint.pformat(request.__dict__, 1, 2400), session=pprint.pformat(dict(session)), message=message, link=link)



#!/usr/bin/env python

import os
import pprint
from setuptools import setup, find_namespace_packages

import logging

logger = logging.getLogger(__name__)

setup_requirements = ["pytest-runner", "setuptools_scm"]

version_file = "./VERSION"
readme_file = "./README.md"


def read_file(fname, strip=True):
    fn = os.path.join(os.path.dirname(os.path.abspath(__file__)), fname)
    data = ""
    if os.path.exists(fn):
        with open(fn) as f:
            data = f.read()
            data = data.strip() if strip else data
            # logger.info(f"Got data '{data}' from '{fn}'")
    else:
        logger.error(f"Could not find file {fn}")
        logger.warning(f"NOTE: Current working directory is {os.getcwd()}")
    return data


def write_file(fname, data, do_overwrite=False):
    fn = os.path.join(os.path.dirname(os.path.abspath(__file__)), fname)
    if not os.path.exists(fn) or do_overwrite:
        with open(fn, "w") as f:
            f.write(data)
    else:
        logger.warning(f"File {fn} already exists")
        logger.warning(f"NOTE: Current working directory is {os.getcwd()}")
    return data


def remove_comment(line, sep="#"):
    i = line.find(sep)
    if i >= 0:
        line = line[:i]
    return line.strip()


def read_requirements_file(fname: str, do_strip: bool = True):
    fn = os.path.join(os.path.dirname(os.path.abspath(__file__)), fname)
    print(f"Reading requirements from {fn} with do_strip = {do_strip}")
    lines = []
    if os.path.exists(fn):
        with open(fn) as f:
            for r in f.readlines():
                r = r.strip()
                if len(r) < 1:
                    continue
                r = remove_comment(r)
                if len(r) < 1:
                    continue
                lines.append(r)
    else:
        logger.error(f"Could not find requirements file {fn}")
        logger.warning(f"NOTE: Current working directory is {os.getcwd()}")
    # logger.warning(f"Full content of '{fname}' was: \n{lines}")
    if not do_strip:
        return lines
    out = []
    for line in lines:
        if line and not line.startswith("-"):
            out.append(line)
    return out


def debug_repo(repo):
    if not repo:
        print(f"No repo")
        return
    print(f"Repository head commit: {repo.head.commit}")
    print(f"Found {len(repo.branches)} branches:")
    for branch in repo.branches:
        print(f" + {branch}({branch.commit})")
    remote = repo.remote()
    print(f"Found {len(remote.refs)} remote refs:")
    for ref in remote.refs:
        print(f" + {ref}({ref.commit})")


def get_git_branch_from_env():
    branch_env = "FK_GIT_ACTUAL_BRANCH"
    branch = os.environ.get(branch_env, None)
    if branch is not None:
        print(f"Using {branch_env} = {branch} from environment")
    else:
        print(f"No value for {branch_env} found")
    return branch


def generate_version_string(version=None, branch=None):
    version = read_file(version_file) if version is None else version
    branch = get_git_branch_from_env() if branch is None else branch
    full_version = ""

    if branch == "production":
        full_version = version
    elif branch == "beta":
        full_version = f"{version}-beta"
    elif branch.startswith("stage-") and len(branch) > 6:
        full_version = f"{version}-{branch[6:]}"
    else:
        full_version = f"{version}-test-{branch.replace(' ','_').replace('	','_')}"
    print(f"Using full version = {full_version}")
    return full_version


def get_version_string():
    # Not viable
    # return generate_version_string();
    return read_file(version_file)


package = {
    "name": "octomy-common",
    "version": get_version_string(),
    "author": "Lennart Rolland",
    "author_email": "lennart@octomy.org",
    "maintainer": "Lennart Rolland",
    "maintainer_email": "lennart@octomy.org",
    "description": ("OctoMY common"),
    "license": "GPL-3 LGPL-3 MIT",
    "platforms": ["Linux"],
    "keywords": "octomy common",
    "url": "https://gitlab.com/octomy/common",
    "packages": find_namespace_packages(
        include=[
            "fk.*",
        ]
    ),
    "zip_safe": True,
    "long_description": read_file(readme_file),
    "long_description_content_type": "text/markdown",
    "setup_requires": setup_requirements,
    "zip_safe": True,
    "install_requires": read_requirements_file("requirements/requirements.in"),  # Allow flexible deps for install
    "tests_require": read_requirements_file("requirements/test_requirements.txt"),  # Use rigid deps for testing
    "test_suite": "tests",
    "python_requires": ">=3.7.4",
    "data_files": [("common", [version_file])],
    "include_package_data": True,
    # From https://pypi.org/pypi?%3Aaction=list_classifiers
    "classifiers": ["Development Status :: 3 - Alpha", "Intended Audience :: Developers", "Intended Audience :: Other Audience", "Topic :: Utilities", "Natural Language :: English", "Operating System :: POSIX :: Linux", "Programming Language :: Python :: 3.7", "Topic :: Other/Nonlisted Topic"],
}

# pprint.pprint(package)
setup(**package)



from pathlib import Path
import pystache


class Svg:
    def __init__(self):
        # Load SVG template from file
        svg_template_fn = "template.svg"
        svg_template_file = Path(svg_template_fn)
        if svg_template_file.is_file():
            with open(svg_template_fn, "r") as svg_template_handle:
                print("Loading svg template from file")
                svg_template_text = svg_template_handle.read()
                # Render the template
                svg_output_fn = "output.svg"
                print("Applying data to template")
                svg_output_text = pystache.render(svg_template_text, {"from": "BGO", "to": "KBP"})
                # Save rendered text into SVG output file
                with open(svg_output_fn, "w") as svg_output_handle:
                    print("Writing output svg")
                    svg_output_handle.write(svg_output_text)
        else:
            print("Template not found, aboring")



import pprint
import logging
import os

from fk.db.DatabaseConnection import DatabaseConnection
from urllib.parse import urlparse, urlunparse, ParseResult

logger = logging.getLogger(__name__)

# fmt: off
configs={
    "empty:":(
        {}
        ,""
        ,""
    )
    , "hello:":(
        {
            "db-hostname":"hello.com"
            ,"db-port":1234
            ,"db-username":"arnold"
            ,"db-password":"secret123"
            ,"db-database":"mydb"
        }
        ,"F12F52B73358C297F47A80768ABDFADF20D021F6A20E9929178908F981B75FA1"
        ,"postgres://arnold:secret123@hello.com:1234/mydb"
    )
}
# fmt: on


def test_true():
    logger.info("Dummy unit test")
    return True


def test_db_get_config_hash():

    for name, pack in configs.items():
        logger.info(f"NAME:{name}")
        config, expected, _ = pack
        logger.info(f"config:{config}")
        logger.info(f"expected:{expected}")
        actual = DatabaseConnection.get_config_hash(config)
        logger.info(f"actual:{actual}")
        assert actual == expected
    return True


def test_db_url_to_and_from_config():
    for name, pack in configs.items():
        logger.info(f"NAME:{name}")
        expected_config, _, expected_url = pack
        logger.info(f"expected_config:{expected_config}")
        logger.info(f"expected_url:   {expected_url}")
        actual_url = DatabaseConnection.config_to_url(expected_config)
        actual_config = DatabaseConnection.url_to_config(expected_url)
        assert actual_url == expected_url
        assert actual_config == expected_config


def _test_db_get_same_twice():
    # fmt: off
    config= {
        "db-hostname":"hello.com"
        ,"db-port":"1234"
        ,"db-username":"arnold"
        ,"db-password":"secret123"
        ,"db-database":"mydb"
    }
    # fmt: on
    db1 = DatabaseConnection.get_connection(config)
    db2 = DatabaseConnection.get_connection(config)
    assert db1 == db2



#!/usr/bin/env python
import os
import posixpath
import hashlib
from flask import current_app
import logging
import fk

logger = logging.getLogger(__name__)


# from https://github.com/pallets/flask/blob/024f0d384cf5bb65c76ac59f8ddce464b2dc2ca1/src/flask/helpers.py

# what separators does this operating system provide that are not a slash?
# this is used by the send_from_directory function to ensure that nobody is
# able to access files from outside the filesystem.
_os_alt_seps = list(sep for sep in [os.path.sep, os.path.altsep] if sep not in (None, "/"))


def safe_join(directory, *pathnames):
    """Safely join `directory` and zero or more untrusted `pathnames`
    components.
    Example usage::
        @app.route('/wiki/<path:filename>')
        def wiki_page(filename):
            filename = safe_join(app.config['WIKI_FOLDER'], filename)
            with open(filename, 'rb') as fd:
                content = fd.read()  # Read and process the file content...
    :param directory: the trusted base directory.
    :param pathnames: the untrusted pathnames relative to that directory.
    :raises: :class:`~werkzeug.exceptions.NotFound` if one or more passed
            paths fall out of its boundaries.
    """

    parts = [directory]

    for fn in pathnames:
        if fn != "":
            pre = fn
            fn = posixpath.normpath(fn)
            # logger.info(f"fn={fn}, pre={pre}")
        if any(sep in fn for sep in _os_alt_seps) or os.path.isabs(fn) or fn == ".." or fn.startswith("../"):
            # logger.info(f"pathname was invalid: {fn}")
            return None

        parts.append(fn)

    return posixpath.join(*parts)


expiry_cache = {}


def get_expiry(path_raw):
    direc = current_app.config["static-files-root"]
    static_prefix = "/static/"
    path = path_raw[len(static_prefix) :] if path_raw.startswith(static_prefix) else path_raw
    filename_temp = os.fspath(path)
    directory = os.fspath(direc)
    filename = safe_join(directory, filename_temp)
    # logger.info(f"path={path}, direc={direc}, filename_temp={filename_temp}, directory={directory}, filename={filename}")
    if not filename:
        # logger.info("No filename returned from safejoin")
        return path_raw
    if not os.path.isabs(filename):
        # logger.info(f"{filename} was not abolute")
        filename = os.path.join(current_app.root_path, filename)
    try:
        if os.path.isfile(filename):
            if not filename in expiry_cache:
                m = hashlib.sha256(fk.utils.read_file(filename).encode("utf-8"))
                expiry_cache[filename] = m.hexdigest()[0:6]
            return f"{path_raw}?expiry_hash={expiry_cache[filename]}"
        else:
            # logger.info(f"filename={filename} was not afile")
            pass
    except Exception as e:
        # logger.info(f"Exception:{e}", exc_info=True)
        pass
    return path_raw
